//
// Copyright (C) 2001 Simon Greenwold. You are allowed to use or modify the code, but this note must be kept intact.
// 
// AgentControl.cpp: implementation of the AgentControl class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include <maya/MGlobal.h>
#include "MFCPlugin.h"
#include "ZoneTypes.h"
#include "AgentControl.h"
#include "Message.h"
#include "AgentLocalizedBoundingBox.h"
#include "AgentOverallDesignHeight.h"
#include "AgentBoundingBox.h"
#include "AgentShapeSize.h"
#include "AgentNumShapes.h"
#include "AgentType.h"
#include "PopulationProperties.h"
#include <iostream>

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

// Declare static members
list<Agent*> AgentControl::activeAgents;
list<Agent*> AgentControl::unfinishedAgents;
Design* AgentControl::currentDesign = NULL;
AgentControl*	AgentControl::theAgentControl;
list<AgentType*> AgentControl::agentTypes;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////


AgentControl::AgentControl(Design* startingDesign)
{
	currentDesign = startingDesign;
	agentTypes.clear();
	activeAgents.clear();
	unfinishedAgents.clear();
	instantiateAgents();
	theAgentControl = this;
}

AgentControl::~AgentControl()
{

}

void AgentControl::instantiateAgents() {	// Create all your agents here
											// DO NOT add them to active agents; that is automatic
											// Do NOT create multiple instances. That is automatic
											// Just initialize each kind once.
	
	Agent* tempAgent;

	agentTypes.push_back((tempAgent = new(AgentOverallDesignHeight))->makeType());
	delete tempAgent;
	agentTypes.push_back((tempAgent = new(AgentLocalizedBoundingBox))->makeType());
	delete tempAgent;
	agentTypes.push_back((tempAgent = new(AgentBoundingBox))->makeType());
	delete tempAgent;
	agentTypes.push_back((tempAgent = new(AgentShapeSize))->makeType());
	delete tempAgent;
	agentTypes.push_back((tempAgent = new(AgentNumShapes))->makeType());
	delete tempAgent;


}

double AgentControl::evaluate(Design* theDesign) {
	currentDesign = theDesign;
	unfinishedAgents.clear();
	for (list<Agent*>::iterator i = activeAgents.begin(); i != activeAgents.end(); i++) {
		unfinishedAgents.push_back(*i);
		(*i)->placeDown();
	}

	int stepNum = 1;

	while(unfinishedAgents.size() > 0) {
		i = unfinishedAgents.begin(); 
		while (i != unfinishedAgents.end()) {
			if (! (*i)->step(stepNum)) {			// Agent is done
				i = unfinishedAgents.erase(i);		// remove it from the list
			}
			else i++;
		}
	}
	
	double result = 0.0;
	double totalWeight = 0.0;
	double tempWeight;

	for (list<AgentType*>::iterator j = agentTypes.begin(); j != agentTypes.end(); j++) {
		result += (*j)->accumulate(tempWeight);
		totalWeight += tempWeight;
	}

	if (totalWeight != 0)
		result /= totalWeight;

	theDesign->getPhenome().initializeImpressionMatrix();
	storeImpressions(theDesign->getPhenome());
	theDesign->getPhenome().makeImpressionsValid();
	return result;
}

Agent*	AgentControl::getAgent(AgentID	ID) {
	return (Agent*)ID;
}

void	AgentControl::newAgent(Agent*	theAgent) {		// This is called automatically by Agent's constuctor. Don't call it manually.
	activeAgents.push_back(theAgent);
}

void	AgentControl::removeAgent(Agent*	theAgent) {		// This is called automatically by Agent's destructor. Don't call it manually.
	activeAgents.remove(theAgent);
}

Design*	AgentControl::getCurrentDesign() {
	return currentDesign;
}

void	AgentControl::relay(Message* theMessage) {
	for (list<Agent*>::iterator i = activeAgents.begin(); i != activeAgents.end(); i++)
		if (theMessage->isTo(*i)) (*i)->receiveMessage(theMessage);
}

void	AgentControl::getAgentTypes(int zoneType, list<AgentType*>& typeList)
{
	for (list<AgentType*>::iterator i = agentTypes.begin(); i != agentTypes.end(); i++)
		if (((zoneType >= 0) && (*i)->getPerZoneType()) || ((zoneType < 0) && !((*i)->getPerZoneType()))) typeList.push_back(*i);
}


void AgentControl::setNumZoneTypes(int newNumTypes)
{
	for (list<AgentType*>::iterator i = agentTypes.begin(); i != agentTypes.end(); i++)
		(*i)->setNumZoneTypes(newNumTypes, false);
}

void AgentControl::removeZoneType(int which)
{
	for (list<AgentType*>::iterator i = agentTypes.begin(); i != agentTypes.end(); i++)
		(*i)->removeZoneType(which);
}

void AgentControl::addZoneType(int after)
{
	for (list<AgentType*>::iterator i = agentTypes.begin(); i != agentTypes.end(); i++)
		(*i)->addZoneType(after);
}

double AgentControl::getImpression(bool &valid, int zoneType, AgentType* agent)
{
	int agentTypeNum = 0;
	double imp;
	for (list<AgentType*>::iterator i = agentTypes.begin(); i != agentTypes.end(); i++, agentTypeNum++) {
		if ((*i)->getName()==agent->getName()) {
			if((*i)->getPerZoneType())
				valid = PopulationProperties::thePopPropsPage->getCurrentDesign()->getPhenome().getImpression(agentTypeNum, zoneType, imp);
			else
				valid = PopulationProperties::thePopPropsPage->getCurrentDesign()->getPhenome().getImpression(agentTypeNum, 0, imp);
			return imp;
		}
	}
	valid = false;
	return -1;
}

int AgentControl::getTotalAgentTypes()
{
	return agentTypes.size();
}

void AgentControl::storeImpressions(Phenotype& phenotype)
{
	int agentTypeNum = 0;
	double impression;
	bool valid;

	for (list<AgentType*>::iterator i = agentTypes.begin(); i != agentTypes.end(); i++, agentTypeNum++) {
		if ((*i)->getPerZoneType()) {
			for (int zoneTypeNum = 0; zoneTypeNum < ZoneTypes::zoneTypes->getNumZoneTypes(); zoneTypeNum++) {
				impression = (*i)->getZoneTypeImpression(zoneTypeNum, valid);
				if (valid)
					phenotype.storeImpression(agentTypeNum, zoneTypeNum, impression);
				else
					phenotype.storeImpression(agentTypeNum, zoneTypeNum, -1);
			}
		}
		else {
			impression = (*i)->getZoneTypeImpression(-1, valid);
			if (valid)
				phenotype.storeImpression(agentTypeNum, 0, impression);
			else
				phenotype.storeImpression(agentTypeNum, 0, -1);
		}
	}
}
