//
// Copyright (C) 2001 Simon Greenwold. You are allowed to use or modify the code, but this note must be kept intact.
// 
// Design.cpp: implementation of the Design class.
//
//////////////////////////////////////////////////////////////////////
#include "stdafx.h"
#include <maya/MGlobal.h>
#include "Design.h"
#include "GeneticUtility.h"
#include "Constants.h"
#include "Genotype.h"
#include "AgentControl.h"
#include "PrimaryShape.h"
#include <iostream>
#include <sstream>
#include "MayaUtilities.h"
#include "MELUtilities.h"

#include "AnalysisBoundingBox.h"
#include "AnalysisShapeSize.h"
#include "AnalysisCountShapes.h"
#include "AnalysisOverallDesignHeight.h"

#include "PopulationProperties.h"

// Declare static members

bool	Design::designAnalysesInstantiated = false;
list<Analysis*>	Design::designAnalyses;

// Deal with Representation baggage

Analysis*	Design::getAnalysis(const string& name) {

	for (list<Analysis*>::iterator i = designAnalyses.begin(); i != designAnalyses.end(); i++)
		if ((*i)->getName() == name) return *i;

	return NULL;
}


void	Design::instantiateAnalyses() {
	// Instantiate your analyses here, though
	// Also be sure to add all analyses to the Design's list of them

	designAnalyses.clear();

	designAnalyses.push_front(new(AnalysisBoundingBox));

	designAnalyses.push_front(new(AnalysisShapeSize));

	designAnalyses.push_front(new(AnalysisCountShapes));

	designAnalyses.push_front(new(AnalysisOverallDesignHeight));

	designAnalysesInstantiated = true;
}

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

Design::Design()
{

	fitnessUpdated = false;

	genome = new Genotype();
	phenome = NULL;

	fitnessUpdated = false;

	if (!designAnalysesInstantiated)
		instantiateAnalyses();
	designAnalysesInstantiated = true;

}	

Design::Design(const std::vector<MayaInfo*>& originalCurves, const std::vector<string>& layerNames) {

	fitnessUpdated = false;

	genome = new Genotype(originalCurves, layerNames);
	phenome = NULL;

	fitnessUpdated = false;

	if (!designAnalysesInstantiated) instantiateAnalyses();
	designAnalysesInstantiated = true;

}

Design::Design(const Design &oldDesign) {
	rawFitness = oldDesign.rawFitness;
	standardizedFitness = oldDesign.standardizedFitness;
	adjustedFitness = oldDesign.adjustedFitness;
	
	genome = new Genotype(*oldDesign.genome);
	phenome = NULL;
	fitnessUpdated = oldDesign.fitnessUpdated;

}

Design::~Design()
{
	delete genome;
	if (phenome) delete phenome;
}


void Design::setRawFitness(double newRawFitness)  // To be changed later
{
	rawFitness = newRawFitness;
	standardizedFitness = newRawFitness;
	adjustedFitness = newRawFitness;
}

void Design::mutateParam(ParamType typeOfParam) {
	genome->mutateParam(typeOfParam);
	if (phenome) phenome->needsUpdate();
	fitnessUpdated = false;
}

void Design::mutateInsertion(InsertionType typeOfInsertion) {
	genome->mutateInsertion(typeOfInsertion);
	if (phenome) phenome->needsUpdate();
	fitnessUpdated = false;
}

void Design::mutateDeletion() {
	genome->mutateDeletion();
	if (phenome) phenome->needsUpdate();
	fitnessUpdated = false;
}

void Design::mutateChange(ChangeType typeOfChange) {
	genome->mutateChange(typeOfChange);
	if (phenome) phenome->needsUpdate();
	fitnessUpdated = false;
}

void Design::crossover(int numberOfCrossingPoints, Design* crossoverTemplate) {
	genome->crossover(numberOfCrossingPoints, crossoverTemplate);
	if (phenome) phenome->needsUpdate();
	fitnessUpdated = false;
}

void Design::randomize(int minNumberOfStartingTransformations, int maxNumberOfStartingTransformations) { //create a random list of transformations and random non-transform data
	genome->randomize(minNumberOfStartingTransformations, maxNumberOfStartingTransformations);
	if (phenome) phenome->needsUpdate();
	fitnessUpdated = false;
}

Phenotype& Design::getPhenome()
{
	if(phenome) {
		if (phenome->updated()) return *phenome;
		return interpret();
	}
	else	{
		return interpret();
	}
}

void	Design::setFitnessUpdated(bool updated)
{
	fitnessUpdated = updated;
}

double	Design::getRawFitness() {
	if (fitnessUpdated) return rawFitness;
	calculateFitness(false);
	return rawFitness;
}
		
double	Design::getStandardizedFitness() {
	if (fitnessUpdated) return standardizedFitness;
	calculateFitness(false);
	//Jason - do stuff to get standardized fitness from raw fitness
	return standardizedFitness;
}

double	Design::getAdjustedFitness() {
	if (fitnessUpdated) return adjustedFitness;
	calculateFitness(false);
	//Jason - do stuff to get adjusted fitness from raw or standardized fitness
	return adjustedFitness;
}

const Design& Design::operator=(const Design& oldDesign) {

	delete genome;
	if (phenome) {
		delete phenome;
		phenome = NULL;
	}

	genome = new Genotype(*oldDesign.getGenome());
	
	rawFitness = oldDesign.rawFitness;
	standardizedFitness = oldDesign.standardizedFitness;
	adjustedFitness = oldDesign.adjustedFitness;

	fitnessUpdated = oldDesign.fitnessUpdated;
	
	return oldDesign;
}

Phenotype& Design::interpret(bool startFromZero)
{
	if (phenome) {
		if (phenome->updated()) return *phenome;
	}
	else phenome = new Phenotype(genome);

	if (PopulationProperties::thePopPropsPage->taskUnderway())
		PopulationProperties::thePopPropsPage->step();

	return phenome->interpret(startFromZero);
}

void Design::saveDesign(string& layerName)
{
	string oldLayer;
	getCurrentLayer(oldLayer);
	selectAllOnLayer(oldLayer);
	newLayer(layerName);
	setLayerVisibility(layerName, false);
	setCurrentLayer(layerName);

	eval("duplicate");
	
	//Design*	newDes = new Design(*this);
	//newDes->interpret();
	//newDes->setEraseOnDelete(false);
	//delete newDes;

	setCurrentLayer(oldLayer);
}


void	Design::erase() {
	if (phenome) phenome->erase();
}

double Design::calculateFitness(bool eraseIt)
{
	if(fitnessUpdated) return rawFitness;
	if(phenome) {
		if (!phenome->updated()) interpret();
	}
	else interpret();

	rawFitness = AgentControl::evaluate(this);
	adjustedFitness = rawFitness;
	standardizedFitness = rawFitness;

	fitnessUpdated = true;
	if (eraseIt)	erase();
	phenome->hideLayer();

	return rawFitness;
}

void Design::setEraseOnDelete(bool newErase) {
	if (!phenome)	phenome = new Phenotype(genome);
	
	phenome->setEraseOnDelete(newErase);
}

bool Design::isPhenomeUpdated() const
{
	if (phenome) return phenome->updated();
	else return false;
}


void Design::reduceNumZoneTypes(int newNumZoneTypes)
{
	if (phenome) phenome->needsUpdate();
	fitnessUpdated = false;

	genome->reduceNumZoneTypes(newNumZoneTypes);
}
