//
// Copyright (C) 2001 Simon Greenwold. You are allowed to use or modify the code, but this note must be kept intact.
// 
#include "stdafx.h"
#include <maya/MGlobal.h>
#include <cstdlib>
#include "GeneticEngine.h"
//#include "Plugin.cpp"
#include "GeneticUtility.h"
#include "Population.h"
#include "Random.h"
#include "Constants.h"
#include <algorithm>
#include <iostream>
#include <sstream>
#include <math.h>
#include "Rotation.h"
#include "Scale.h"
#include "Translation.h"
#include "Design.h"
#include "ProgressBar.h"
#include "PopulationProperties.h"
#include "Settings.h"
#include "ZoneTypes.h"
#include <maya/MSelectionList.h>
#include <maya/MItSelectionList.h>
#include "Layer.h"
#include "MELUtilities.h"


// Declare class' statics:
int GeneticEngine::translationPercent;
int GeneticEngine::rotationPercent;
int GeneticEngine::scalePercent;
int GeneticEngine::reflectionPercent;
int GeneticEngine::openingPercent;
double GeneticEngine::ZWideningMin;
double GeneticEngine::ZWideningMax;
double GeneticEngine::ZHeightMin;
double GeneticEngine::ZHeightMax;
double GeneticEngine::ZStartMin;
double GeneticEngine::ZStartMax;
int GeneticEngine::minCrossoverSize;
int GeneticEngine::maxCrossoverSize;

GeneticEngine* GeneticEngine::geneticEngine;

class CompareFitness {
public:
	int operator() (Design* d1, Design* d2)
	{return (d1->getRawFitness() < d2->getRawFitness());}
};

GeneticEngine::GeneticEngine() {

	setGeneticEngineParametersToConstants();
	pop = new Population(populationSize);
	elite = new std::vector<Design*>(elitism);
	
	_selectionList = NULL;

	//seedInitialPopulation();
	numberOfGenerationsRun = 0;
	geneticEngine = this;
}

/*GeneticEngine::GeneticEngine(int newPopulationSize) {  // Don't use this. It has no meaning to create a population with no shapes

	setGeneticEngineParametersToConstants();
	populationSize = newPopulationSize;
	pop = new Population(populationSize, 0);		// Creating designs with no shapes in them
	elite = new std::vector<Design*>(elitism);
	
	seedInitialPopulation();
	numberOfGenerationsRun = 0;
}*/

GeneticEngine::GeneticEngine(int newPopulationSize, const std::vector<string>& shapeNames, const std::vector<string>& layerNames, MSelectionList* selectionList) {

	setGeneticEngineParametersToConstants();
	populationSize = newPopulationSize;
	pop = new Population(populationSize, shapeNames, layerNames);
	elite = new std::vector<Design*>(elitism);

	_selectionList = selectionList;
	
	//seedInitialPopulation();
	numberOfGenerationsRun = 0;
	geneticEngine = this;
}

GeneticEngine::GeneticEngine(Population* newPop) 
{
	setGeneticEngineParametersToConstants();
	setPopulation(newPop);
	elite = new std::vector<Design*>(elitism);

	numberOfGenerationsRun = 0;
	geneticEngine = this;
}

GeneticEngine::~GeneticEngine() {
	delete pop;
	delete elite;
}

void GeneticEngine::clearElite() {

	for(int i = 0; i < elite->size(); i++)
		(*elite)[i] = NULL;
}

void GeneticEngine::setGeneticEngineParametersToConstants() {
	populationSize = POPULATION_SIZE;
	minNumberOfStartingTransformations = MINNUMBEROFSTARTINGTRANSFORMATIONS;
	maxNumberOfStartingTransformations = MAXNUMBEROFSTARTINGTRANSFORMATIONS;

	tournamentSize = TOURNAMENTSIZE;

	paramMutationPercent = PARAMMUTATIONPERCENT;
	PETypePercent = PETYPEPERCENT;
	ZWideningPercent = ZWIDENINGPERCENT;
	shapeModePercent = SHAPEMODEPERCENT;
	ZHeightPercent = ZHEIGHTPERCENT;
	ZStartPercent = ZSTARTPERCENT;
	
	insertionMutationPercent = INSERTIONMUTATIONPERCENT;
	middlePercent = MIDDLEPERCENT;
	endPercent = ENDPERCENT;
	
	changeMutationPercent = CHANGEMUTATIONPERCENT;
	ringPercent = RINGPERCENT;
	paramPercent = PARAMPERCENT;
	completePercent = COMPLETEPERCENT;

	deletePercent = DELETEPERCENT;

	crossoverPercent = CROSSOVERPERCENT;
	minNumberOfCrossingPoints = MINNUMBEROFCROSSINGPOINTS;
	maxNumberOfCrossingPoints = MAXNUMBEROFCROSSINGPOINTS;
	minCrossoverSize = MINCROSSOVERSIZE;
	maxCrossoverSize = MAXCROSSOVERSIZE;
	
	reproductionPercent = REPRODUCTIONPERCENT;
	elitism = ELITISM;
	eliteCopies = ELITECOPIES;

	translationPercent = TRANSLATION_PERCENT;
	rotationPercent = ROTATION_PERCENT;
	scalePercent = SCALE_PERCENT;
	reflectionPercent = REFLECTION_PERCENT;
	openingPercent = OPENING_PERCENT;

	// Maxes and Mins
	ZWideningMin = Z_WIDENING_MIN;
	ZWideningMax = Z_WIDENING_MAX;

	// ZHeight constants
	ZHeightMin = Z_HEIGHT_MIN;
	ZHeightMax = Z_HEIGHT_MAX;

	// ZStart constants
	ZStartMin = Z_START_MIN;
	ZStartMax = Z_START_MAX;

	// Parameter scaling constants for rotation

	Rotation::setMaxRotation(ROTATION_PARAMETER_MAX);
	Rotation::setMinRotation(ROTATION_PARAMETER_MIN);

	// Parameter scaling constants for scaling

	Scale::setMaxXScale(SCALING_PARAMETER_X_MAX);
	Scale::setMinXScale(SCALING_PARAMETER_X_MIN);
	Scale::setMaxYScale(SCALING_PARAMETER_Y_MAX);
	Scale::setMinYScale(SCALING_PARAMETER_Y_MIN);
	Scale::setMaxZScale(SCALING_PARAMETER_Z_MAX);
	Scale::setMinZScale(SCALING_PARAMETER_Z_MIN);

	// Parameter scaling constants for translation

	Translation::setMaxXTranslation(TRANSLATION_PARAMETER_X_MAX);
	Translation::setMinXTranslation(TRANSLATION_PARAMETER_X_MIN);
	Translation::setMaxYTranslation(TRANSLATION_PARAMETER_Y_MAX);
	Translation::setMinYTranslation(TRANSLATION_PARAMETER_Y_MIN);
	Translation::setMaxZTranslation(TRANSLATION_PARAMETER_Z_MAX);
	Translation::setMinZTranslation(TRANSLATION_PARAMETER_Z_MIN);
}

double GeneticEngine::getMeanFitness() {		// Caching could make this more efficient if it ever becomes an issue

	double accumulator = 0.0;

//	popPropsPage->newTask("Evaluating new population...", populationSize);

	for (int i = 0; i < populationSize; i++) {
		accumulator += pop->getDesign(i)->getRawFitness();
	}

	int result = 0;
	
//	popPropsPage->taskFinished();
	return accumulator / populationSize;
}

double GeneticEngine::getStdDevFitness() {		// Caching could make this more efficient if it ever becomes an issue

	double mean = getMeanFitness();
	double accumulator = 0.0;
	double temp = 0.0;

	for (int i = 0; i < populationSize; i++) {
		temp = pop->getDesign(i)->getRawFitness() - mean;
		accumulator += temp * temp;
	}
	return sqrt(accumulator / (double)(populationSize - 1));
}

void GeneticEngine::setPopulation(Population* newPop) {
	delete pop;
	pop = newPop;

	populationSize = pop->getDesigns()->size();
}

void GeneticEngine::seedInitialPopulation() {
	//geneticEngine = initializeEngine();
	pop->randomize(minNumberOfStartingTransformations, maxNumberOfStartingTransformations);
}

int GeneticEngine::tournamentSelect(int firstN, int tournamentSize) {   // Does tournament selection on the first N members of the current population
	int bestDesign, tempDesign;											// The rest have already been moved to the new population
	double bestFitness, tempFitness;									// It returns a reference into the existing population

	bestDesign = safeRandom(firstN - 1);
	bestFitness = pop->getDesign(bestDesign)->getRawFitness();			// fitness updated if necessary

	tournamentSize--;
	while (tournamentSize > 0) {
		tempDesign = safeRandom(firstN - 1);
		tempFitness = pop->getDesign(tempDesign)->getRawFitness();		// fitness updated if necessary
		if(bestFitness > tempFitness) {									//as per code elsewhere, lower fitness is better
			bestDesign = tempDesign;
			bestFitness = tempFitness;
		}
		tournamentSize--;
	}
	return bestDesign;
}



void GeneticEngine::breedPopulation() {

	
	int individualsLeft = populationSize;
	double currentFitness;
	std::vector<Design*>* newPop = new std::vector<Design*>(populationSize);  // The new population

	// Select elite members to next population
	clearElite();
	vector<double> eliteFitnesses(elitism);
	for (int i = 0; i < elitism; i++)
		eliteFitnesses[i] = 1000000;   // If we have fitness higher than this, we have some other bug (remember, low is good)

	for (i = 0; i < populationSize; i++) {
		currentFitness = pop->getDesign(i)->getRawFitness();
		for (int j = 0; j < elitism; j++) {
			if (currentFitness < eliteFitnesses[j]) {
				for (int k = j + 1; k < elitism; k++) {
					eliteFitnesses[k] = eliteFitnesses[k-1];
					(*elite)[k] = (*elite)[k-1];
				}
				eliteFitnesses[j] = currentFitness;
				(*elite)[j] = pop->getDesign(i);

				j = elitism;
			}
		}
	}

	
	//Copy elites
	for (i = 0; i < elite->size(); i++) {
		if ((*elite)[i]) {
			for (int j = 0; j < eliteCopies; j++) {
				(*newPop)[individualsLeft - 1] = new Design(*(*elite)[i]);	// Copy design pointer to new population
				(*newPop)[individualsLeft - 1]->setFitnessUpdated(false);
				individualsLeft--;
			}

		}
		else
			cout << "Elitism error.";
	}

	//select the rest of the next generation's population and perform mutation and crossover
	int breedType;
	Design* currentDesign;

	while (individualsLeft > 0) {

		// First copy existing design into the new population
		currentDesign = new Design (*pop->getDesign(tournamentSelect(populationSize, tournamentSize))); //currentDesign will be modified directly and brought over to the new population
		currentDesign->setFitnessUpdated(false);

		int selectType;
		int numberOfCrossingPoints;
		int numMutations;
		if ((getMeanFitness() != 0.0) && (fabs(getStdDevFitness() / getMeanFitness()) <= 2.0)) {
			numMutations = safeRandom(fabs(1.0 - (getStdDevFitness() / getMeanFitness())) * MUTATION_RATE) + 1;
		}
		else
			numMutations = 1;

		for (i = 0; i < numMutations; i++) {
			breedType = safeRandom(99);
			if(breedType < paramMutationPercent) {				// SHAPE PARAMETER MUTATION
				ParamType paramType;
				selectType = safeRandom(99);
				if(selectType < PETypePercent) { 
					paramType = PETYPE;
					numPETyped++;
				}
				else if((selectType -= PETypePercent) < ZWideningPercent) {
					paramType = ZWIDENING;
					numZWidened++;
				}
				else if((selectType -= ZWideningPercent) < shapeModePercent) {
					paramType = SHAPEMODE;
					numshapeModed++;
				}
				else if((selectType -= shapeModePercent) < ZHeightPercent) {
					paramType = ZHEIGHT;
					numZHeighted++;
				}
				else {
					paramType = ZSTART;
					numZStarted++;
				}

				currentDesign->mutateParam(paramType);
				numParamMutated++;

			}
			else if((breedType -= paramMutationPercent) < insertionMutationPercent) {		// INSERTION COMMAND MUTATION
				InsertionType insertionType;
				selectType = safeRandom(99);
				if(selectType < middlePercent) { 
					insertionType = MIDDLE;
					numMiddled++;
				}
				else {
					insertionType = END;
					numEnded++;
				}

				currentDesign->mutateInsertion(insertionType);
				numInsertionMutated++;
			}
			else if((breedType -= insertionMutationPercent) < changeMutationPercent) {           // CHANGE EXISTING COMMAND MUTATION
				ChangeType changeType;												// (RING, PARAM, TYPE, COMPLETE)
				selectType = safeRandom(99);

				if(selectType < ringPercent) {
					changeType = RING;
					numRinged++;
				}
				else if((selectType -= ringPercent) < paramPercent) {
					changeType = PARAM;
					numParamed++;
				}
				else {
					changeType = COMPLETE;
					numCompleted++;
				}

				currentDesign->mutateChange(changeType);
				numChangeMutated++;
			}
			else if((breedType -= changeMutationPercent) < crossoverPercent) {				// CROSSOVER

				numberOfCrossingPoints = randomRange(minNumberOfCrossingPoints, maxNumberOfCrossingPoints);
				currentDesign->crossover(numberOfCrossingPoints, pop->getDesign(tournamentSelect(populationSize, tournamentSize)));
				numCrossedOver++;
			}
			else if ((breedType -= crossoverPercent) < deletePercent) {
				currentDesign->mutateDeletion();
				numDeleted++;
			}
			else
				numReproduced++;
			//else reproduction mutation - do not change currentDesign
		}

		(*newPop)[individualsLeft - 1] = currentDesign;

		individualsLeft--;
	}
	
	std::sort(newPop->begin(), newPop->end(), CompareFitness());	// Sort the population by fitness for convenience
	pop->setDesigns(newPop);
}



Population* GeneticEngine::runPopulation(int numberOfGenerations) 
{
	for(int i = 0; i < numberOfGenerations; i++) {
		clearStats();
		breedPopulation();
		reportStats();
		numberOfGenerationsRun++;
	}
	
	return pop;
}

void GeneticEngine::setPopulationSize(int newSize) {
	if (newSize < 1 || newSize == populationSize) return;
	
	if(PopulationProperties::thePopPropsPage->isInitialized())	{
		PopulationProperties::thePopPropsPage->changeCurrent(0);
		int i;
		Design *newDesign;

		if (newSize > populationSize) {
			for (i = 0; i < newSize - populationSize; i++) {
				newDesign = new Design(*(pop->getDesign(safeRandom(populationSize))));
				pop->getDesigns()->push_back(newDesign);
				//delete newDesign;
			}
		}
		else {													// new size is smaller than old
			for (i = 0; i < populationSize - newSize; i++) {
				pop->getDesign(populationSize - (i + 1))->erase();
				delete pop->getDesign(populationSize - (i + 1));
				pop->getDesigns()->erase(pop->getDesigns()->end() - 1);
			}
		}
	}
	else	{
		//geneticEngine = initializeEngine();
		int i;
		Design *newDesign;

		if (newSize > populationSize) {
			Sleep(1);
			for (i = 0; i < newSize - populationSize; i++) {
				newDesign = new Design(*(pop->getDesign(safeRandom(populationSize))));
				pop->getDesigns()->push_back(newDesign);
				//delete newDesign;
			}
		}
		else {													// new size is smaller than old
			for (i = 0; i < populationSize - newSize; i++) {
				pop->getDesign(populationSize - (i + 1))->erase();
				delete pop->getDesign(populationSize - (i + 1));
				pop->getDesigns()->erase(pop->getDesigns()->end() - 1);
			}
		}
	}
	populationSize = newSize;
}

void GeneticEngine::clearStats() {

	numParamMutated = 0;
	numPETyped = 0;
	numZWidened = 0;
	numshapeModed = 0;
	numZHeighted = 0;
	numZStarted = 0;

	numInsertionMutated = 0;
	numMiddled = 0;
	numEnded = 0;

	numChangeMutated = 0;
	numRinged = 0;
	numParamed = 0;
	numCompleted = 0;

	numCrossedOver = 0;
	numReproduced = 0;
	numDeleted = 0;
}

void GeneticEngine::reportStats() {

	cout << "Population size: " << populationSize << endl;
	cout << "Parameter mutations: " << numParamMutated << endl;
	cout << "  PEType mutations:     " << numPETyped << endl;
	cout << "  ZWidening mutations:  " << numZWidened << endl;
	cout << "  Shape mode mutations: " << numshapeModed << endl;
	cout << "  ZHeight mutations:    " << numZHeighted << endl;
	cout << "  ZStart mutations:     " << numZStarted << endl;
	cout << "Insertion mutations: " << numInsertionMutated << endl;
	cout << "  Middle insertions: " << numMiddled << endl;
	cout << "  End insertions:    " << numEnded << endl;
	cout << "Command change mutations: " << numChangeMutated << endl;
//	cout << "  Ring changes:       " << numRinged << endl;
	cout << "  Parameter changes:  " << numParamed << endl;
	cout << "  Complete changes:   " << numCompleted << endl;
	cout << "Crossover mutations: " << numCrossedOver << endl;
	cout << "Reproductions: " << numReproduced << endl;
	cout << "Command deletions: " << numDeleted << endl;

	cout << endl;
	cout << endl;

}

void	GeneticEngine::returnStats(std::string& stats) {

	ostringstream tempStr;

	tempStr << "\r\n";

	tempStr << "Generation: " << numberOfGenerationsRun << "\r\n";
	tempStr << "Parameter mutations: " << numParamMutated << "\r\n";
	tempStr << "  PEType mutations:     " << numPETyped << "\r\n";
	tempStr << "  ZWidening mutations:  " << numZWidened << "\r\n";
	tempStr << "  Shape mode mutations: " << numshapeModed << "\r\n";
	tempStr << "  ZHeight mutations:    " << numZHeighted << "\r\n";
	tempStr << "  ZStart mutations:     " << numZStarted << "\r\n";
	tempStr << "Insertion mutations: " << numInsertionMutated << "\r\n";
	tempStr << "  Middle insertions: " << numMiddled << "\r\n";
	tempStr << "  End insertions:    " << numEnded << "\r\n";
	tempStr << "Command change mutations: " << numChangeMutated << "\r\n";
//	tempStr << "  Ring changes:       " << numRinged << "\r\n";
	tempStr << "  Parameter changes:  " << numParamed << "\r\n";
	tempStr << "  Complete changes:   " << numCompleted << "\r\n";
	tempStr << "Crossover mutations: " << numCrossedOver << "\r\n";
	tempStr << "Reproductions: " << numReproduced << "\r\n";
	tempStr << "Command deletions: " << numDeleted << "\r\n";

	stats = tempStr.str();
}


void GeneticEngine::reportSettings()
{
	Settings::settings->setSetting("GeneticEngine", "PopSize", populationSize);
	Settings::settings->setSetting("GeneticEngine", "TournamentSize", tournamentSize);
	Settings::settings->setSetting("GeneticEngine", "Elites", elitism);
	Settings::settings->setSetting("GeneticEngine", "EliteCopies", eliteCopies);
	Settings::settings->setSetting("GeneticEngine", "paramMutationPercent", paramMutationPercent);
	Settings::settings->setSetting("GeneticEngine", "PETypePercent", PETypePercent);
	Settings::settings->setSetting("GeneticEngine", "ZWideningPercent", ZWideningPercent);
	Settings::settings->setSetting("GeneticEngine", "shapeModePercent", shapeModePercent);
	Settings::settings->setSetting("GeneticEngine", "ZHeightPercent", ZHeightPercent);
	Settings::settings->setSetting("GeneticEngine", "ZStartPercent", ZStartPercent);
	Settings::settings->setSetting("GeneticEngine", "insertionMutationPercent", insertionMutationPercent);
	Settings::settings->setSetting("GeneticEngine", "middlePercent", middlePercent);
	Settings::settings->setSetting("GeneticEngine", "endPercent", endPercent);
	Settings::settings->setSetting("GeneticEngine", "changeMutationPercent", changeMutationPercent);
	Settings::settings->setSetting("GeneticEngine", "ringPercent", ringPercent);
	Settings::settings->setSetting("GeneticEngine", "paramPercent", paramPercent);
	Settings::settings->setSetting("GeneticEngine", "completePercent", completePercent);
	Settings::settings->setSetting("GeneticEngine", "deletePercent", deletePercent);
	Settings::settings->setSetting("GeneticEngine", "crossoverPercent", crossoverPercent);
	Settings::settings->setSetting("GeneticEngine", "reproductionPercent", reproductionPercent);
}

void GeneticEngine::receiveSettings()
{
	
	pop->clearFitnesses();

	int tempInt;
	if (Settings::settings->getSetting("GeneticEngine", "PopSize", tempInt)) {
		setPopulationSize(tempInt);
	}
	else setPopulationSize(POPULATION_SIZE);

	if (Settings::settings->getSetting("GeneticEngine", "TournamentSize", tempInt)) {
		tournamentSize = tempInt;
	}
	else tournamentSize = TOURNAMENTSIZE;

	if (Settings::settings->getSetting("GeneticEngine", "Elites", tempInt)) {
		elitism = tempInt;
	}
	else elitism = ELITISM;

	if (Settings::settings->getSetting("GeneticEngine", "EliteCopies", tempInt)) {
		eliteCopies = tempInt;
	}
	else eliteCopies = ELITECOPIES;

	if (Settings::settings->getSetting("GeneticEngine", "paramMutationPercent", tempInt)) {
		paramMutationPercent = tempInt;
	}
	else paramMutationPercent = PARAMMUTATIONPERCENT;

	if (Settings::settings->getSetting("GeneticEngine", "PETypePercent", tempInt)) {
		PETypePercent = tempInt;
	}
	else PETypePercent = PETYPEPERCENT;

	if (Settings::settings->getSetting("GeneticEngine", "ZWideningPercent", tempInt)) {
		ZWideningPercent = tempInt;
	}
	else ZWideningPercent = ZWIDENINGPERCENT;

	if (Settings::settings->getSetting("GeneticEngine", "shapeModePercent", tempInt)) {
		shapeModePercent = tempInt;
	}
	else shapeModePercent = SHAPEMODEPERCENT;

	if (Settings::settings->getSetting("GeneticEngine", "ZHeightPercent", tempInt)) {
		ZHeightPercent = tempInt;
	}
	else ZHeightPercent = ZHEIGHTPERCENT;

	if (Settings::settings->getSetting("GeneticEngine", "ZStartPercent", tempInt)) {
		ZStartPercent = tempInt;
	}
	else ZStartPercent = ZSTARTPERCENT;

	if (Settings::settings->getSetting("GeneticEngine", "insertionMutationPercent", tempInt)) {
		insertionMutationPercent = tempInt;
	}
	else insertionMutationPercent = INSERTIONMUTATIONPERCENT;

	if (Settings::settings->getSetting("GeneticEngine", "middlePercent", tempInt)) {
		middlePercent = tempInt;
	}
	else middlePercent = MIDDLEPERCENT;

	if (Settings::settings->getSetting("GeneticEngine", "endPercent", tempInt)) {
		endPercent = tempInt;
	}
	else endPercent = ENDPERCENT;

	if (Settings::settings->getSetting("GeneticEngine", "changeMutationPercent", tempInt)) {
		changeMutationPercent = tempInt;
	}
	else changeMutationPercent = CHANGEMUTATIONPERCENT;

	if (Settings::settings->getSetting("GeneticEngine", "ringPercent", tempInt)) {
		ringPercent = tempInt;
	}
	else ringPercent = RINGPERCENT;

	if (Settings::settings->getSetting("GeneticEngine", "paramPercent", tempInt)) {
		paramPercent = tempInt;
	}
	else paramPercent = PARAMPERCENT;

	if (Settings::settings->getSetting("GeneticEngine", "completePercent", tempInt)) {
		completePercent = tempInt;
	}
	else completePercent = COMPLETEPERCENT;

	if (Settings::settings->getSetting("GeneticEngine", "deletePercent", tempInt)) {
		deletePercent = tempInt;
	}
	else deletePercent = DELETEPERCENT;

	if (Settings::settings->getSetting("GeneticEngine", "crossoverPercent", tempInt)) {
		crossoverPercent = tempInt;
	}
	else crossoverPercent = CROSSOVERPERCENT;

	if (Settings::settings->getSetting("GeneticEngine", "reproductionPercent", tempInt)) {
		reproductionPercent = tempInt;
	}
	else reproductionPercent = REPRODUCTIONPERCENT;
}

void GeneticEngine::reduceNumZoneTypes(int newNumZoneTypes)  // 0 - (newNumZoneTypes - 1)
{
	for (int i = 0; i < populationSize; i++) {
		pop->getDesign(i)->reduceNumZoneTypes(newNumZoneTypes);
	}
}

void GeneticEngine::makeAllImpressionsInvalid()
{
	for(int i=0; i < getPopulationSize(); i++)	{
		getPopulation()->getDesign(i)->getPhenome().makeImpressionsInvalid();
	}
}

GeneticEngine* GeneticEngine::initializeEngine() 
{

	GeneticEngine* theEngine = NULL;
	//MSelectionList      list;
	MStringArray		names;
	std::vector<string> shapeNames;
	std::vector<string> layerNames;
	cout << "Initializing engine after changing popsize." << endl;
	int selCount = 0;
    //MGlobal::getActiveSelectionList( (*_selectionList) );
	for ( MItSelectionList listIter( (*_selectionList), MFn::kNurbsCurve ); !listIter.isDone(); listIter.next() )
	{
		cout << "Counting curves." << endl;
		selCount++;
	}

	if (selCount == 0) {
		AfxMessageBox("You must select at least one NURBS curve for AgencyGP to work.");
		return NULL;
	}

// This code gets the Maya curves selected and closes them

	shapeNames.resize(selCount);
	layerNames.resize(selCount);
	int i = 0;
	string layerName;
	for ( listIter.reset(); !listIter.isDone(); listIter.next(), i++)
	{
		listIter.getStrings(names);
		shapeNames[i] = (names[0]).asChar();
		//eval("int $isClosed = `getAttr \"" + shapeNames[i] + ".form\"`");
		//if (getMELint("$isClosed") < 2)
		//	eval("closeCurve -ch off -rpo on " + shapeNames[i]);

		eval("string $layername = `objectLayer " + shapeNames[i] + "`");
		getMELvalue("$layername", layerName);
		layerNames[i] = layerName;
	}

	theEngine = new GeneticEngine(POPULATION_SIZE, shapeNames, layerNames, _selectionList);

	if (theEngine == NULL) AfxMessageBox("Genetic engine could not be initialized.");

	return theEngine;
}
