//
// Copyright (C) 2001 Simon Greenwold. You are allowed to use or modify the code, but this note must be kept intact.
// 
// GeneticEngine.h: interface for the GeneticEngine class.
//
// GeneticEngine contains, aptly enough, the genetic engine
// Specifically, it contains the highest level poceedures for constructing
// a new generation from an old one using genetic operators based upon fitness
//////////////////////////////////////////////////////////////////////

#if !defined(GENETIC_ENGINE_H__INCLUDED_)
//#ifndef GENETIC_ENGINE_H__INCLUDED_
#define GENETIC_ENGINE_H__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "Population.h"
#include "Settable.h"
#include <maya/MItSelectionList.h>
#include <maya/MSelectionList.h>

#include <string>
class Design;
class PopulationProperties;

class GeneticEngine : public Settable {
public:
	GeneticEngine();
//	GeneticEngine(int newPopulationSize); //create a new Genetic Engine with a specified population size
	GeneticEngine(int newPopulationSize, const std::vector<string>& shapeNames, const std::vector<string>& layerNames, MSelectionList* selectionList); //create a new Genetic Engine with a specified population size and primary shape names
	GeneticEngine(Population* newPop); //Create based on an existing population
	virtual ~GeneticEngine();

//Accessors
	Population*	getPopulation() { return pop; }; //returns a reference to the population
	int					getGenerationsRun() { return numberOfGenerationsRun; }; //returns the number of generations run in this (continuous) run
	double				getMeanFitness(); // Return mean fitness
	double				getStdDevFitness(); // Return the standard deviation of fitnesses
	int					getPopulationSize() {return populationSize;}; // Returns population size
	void				returnStats(string& stats); // return all the generation stats

	int	getParamMutationPercent() {return paramMutationPercent;};
	int getPETypePercent() {return PETypePercent;};
	int getZWideningPercent() {return ZWideningPercent;};
	int getShapeModePercent() {return shapeModePercent;};
	int getZHeightPercent() {return ZHeightPercent;};
	int getZStartPercent() {return ZStartPercent;};
	
	int	getInsertionMutationPercent() {return insertionMutationPercent;};
	int getMiddlePercent() {return middlePercent;};
	int getEndPercent() {return endPercent;};
	
	int	getChangeMutationPercent() {return changeMutationPercent;};
	int getRingPercent() {return ringPercent;};
	int getParamPercent() {return paramPercent;};
	int getCompletePercent() {return completePercent;};

	int getDeletePercent() {return deletePercent;};

	int	getCrossoverPercent() {return crossoverPercent;};
	int getMinNumberOfCrossingPoints() {return minNumberOfCrossingPoints;};
	int getMaxNumberOfCrossingPoints() {return maxNumberOfCrossingPoints;};	
	static int getMinCrossoverSize() {return minCrossoverSize;};
	static int getMaxCrossoverSize() {return maxCrossoverSize;};
	
	int	getReproductionPercent() {return reproductionPercent;};
	int	getElitism() {return elitism;};
	int	getEliteCopies() {return eliteCopies;};

	int getTournamentSize() {return tournamentSize;};

	static int getTranslationPercent() {return translationPercent;};
	static int getRotationPercent() {return rotationPercent;};
	static int getScalePercent() {return scalePercent;};
	static int getReflectionPercent() {return reflectionPercent;};
	static int getOpeningPercent() {return openingPercent;};

	static double getZWideningMin() {return ZWideningMin;};
	static double getZWideningMax() {return ZWideningMax;};

	static double getZHeightMin() {return ZHeightMin;};
	static double getZHeightMax() {return ZHeightMax;};

	static double getZStartMin() {return ZStartMin;};
	static double getZStartMax() {return ZStartMax;};

//Mutators

	//Interface
	void	setPopProps(PopulationProperties* newPopPropsPage) {popPropsPage = newPopPropsPage;};

	void				setPopulation(Population* newPop); //sets the population on which to be operated
	Population*			runPopulation(int numberOfGenerations); //runs the given population for a set number of generations - this is cummulative to any prior runPopulation() calls with the current population
	void		setPopulationSize(int newSize); // Sets the population size to a new value. A little complicated.


	void	setParamMutationPercent(int newParamMutationPercent) {paramMutationPercent = newParamMutationPercent;};
	void	setPETypePercent(int newPETypePercent) {PETypePercent = newPETypePercent;};
	void	setZWideningPercent(int newZWideningPercent) {ZWideningPercent = newZWideningPercent;};
	void	setShapeModePercent(int newShapeModePercent) {shapeModePercent = newShapeModePercent;};
	void	setZHeightPercent(int newZHeightPercent) {ZHeightPercent = newZHeightPercent;};
	void	setZStartPercent(int newZStartPercent) {ZStartPercent = newZStartPercent;};
	
	void	setInsertionMutationPercent(int newInsertionMutationPercent) { insertionMutationPercent = newInsertionMutationPercent;};
	void	setMiddlePercent(int newMiddlePercent) {middlePercent = newMiddlePercent;};
	void	setEndPercent(int newEndPercent) {endPercent = newEndPercent;};
	
	void	setChangeMutationPercent(int newChangeMutationPercent) {changeMutationPercent = newChangeMutationPercent;};
	void	setRingPercent(int newRingPercent) {ringPercent = newRingPercent;};
	void	setParamPercent(int newParamPercent) {paramPercent = newParamPercent;};
	void	setCompletePercent(int newCompletePercent) {completePercent = newCompletePercent;};

	void	setDeletePercent(int newDeletePercent) {deletePercent = newDeletePercent;};

	void	setCrossoverPercent(int newCrossoverPercent) {crossoverPercent = newCrossoverPercent;};
	void	setMinNumberOfCrossingPoints(int newMinNumberOfCrossingPoints) {minNumberOfCrossingPoints = newMinNumberOfCrossingPoints;};
	void	setMaxNumberOfCrossingPoints(int newMaxNumberOfCrossingPoints) {maxNumberOfCrossingPoints = newMaxNumberOfCrossingPoints;};	
	static void	setMinCrossoverSize(int newMinCrossoverSize) {minCrossoverSize = newMinCrossoverSize;};
	static void	setMaxCrossoverSize(int newMaxCrossoverSize) {maxCrossoverSize = newMaxCrossoverSize;};
	
	void	setReproductionPercent(int newReproductionPercent) {reproductionPercent = newReproductionPercent;};
	void	setElitism(int newElitism) {elitism = newElitism;};
	void	setEliteCopies(int newEliteCopies) {eliteCopies = newEliteCopies;};

	void	setTournamentSize(int newTournamentSize) {tournamentSize = newTournamentSize;};

	static void setTranslationPercent(int newTranslationPercent) {translationPercent = newTranslationPercent;};
	static void setRotationPercent(int newRotationPercent) {rotationPercent = newRotationPercent;};
	static void setScalePercent(int newScalePercent) {scalePercent = newScalePercent;};
	static void setReflectionPercent(int newReflectionPercent) {reflectionPercent = newReflectionPercent;};
	static void setOpeningPercent(int newOpeningPercent) {openingPercent = newOpeningPercent;};

	static void setZWideningMin(double newZWideningMin) {ZWideningMin = newZWideningMin;};
	static void setZWideningMax(double newZWideningMax) {ZWideningMax = newZWideningMax;};

	static void setZHeightMin(double newZHeightMin) {ZHeightMin = newZHeightMin;};
	static void setZHeightMax(double newZHeightMax) {ZHeightMax = newZHeightMax;};

	static void setZStartMin(double newZStartMin) {ZStartMin = newZStartMin;};
	static void setZStartMax(double newZStartMax) {ZStartMax = newZStartMax;};

	void	reduceNumZoneTypes(int newNumZoneTypes);

	bool	isSeeded()	{return bIsSeeded;};
	void	SeedEngine()	{bIsSeeded = true;};

	void	seedInitialPopulation();						//create a new population by using fitness based (currently tournament) selection to choose individuals to be advanced unchanged, mutated, or crossed over

private:
	void	clearElite();									//Clears out the elite vector at the beginning of each generation
	void	setGeneticEngineParametersToConstants();		//sets the parameters that control the operation of the genetic engine to those constants specified in constants.h
															//The constants are stored in Constants.h instead of in this function so all constants can be collected in a single location, easing recompilation.  Additionally, a function like this allows other functions that set parameters based upon option files to be easily integrated in the future
							//fill the current population with random individuals - use a linear ramp of transfomation list sizes as in lil-gp
	int		tournamentSelect(int firstN, int tournamentSize);	//return the most fit of tournamentSize number of randomly selected individuals from the firstN of the current population
	void	breedPopulation();								//run the current population through one generation 
	

	void	reportStats(); // Print out all the generation stats
	void	clearStats();  // Clear all the generation stats


	Population*	pop;			//the current population on which all functions act by default
	std::vector<Design*>* elite;	// Stores pointers into the population to the top "elitism" designs in the current population

	int			populationSize; //the number of Designs in the current population
	int			numberOfGenerationsRun; //the number of generations through which the current population has been run (consecutively)
	
	int minNumberOfStartingTransformations; //the range of lengths of transformation lists of Designs used in seedInitialPopulation()
	int maxNumberOfStartingTransformations;

	int tournamentSize;

	int	paramMutationPercent; //percent (out of 100) of individuals in a new population that will be modified by a mutation that changes a PrimaryShape's intrinsic parameters rather than its commands
	int PETypePercent; //percent of param mutations that affect PEType
	int ZWideningPercent; //" " ZWidening
	int shapeModePercent; //" " shapeMode (1 - 45 how intersections are handled)
	int ZHeightPercent; //" " ZHeight
	int ZStartPercent; //" " ZStart
	
	int	insertionMutationPercent; //percent (out of 100) of individuals in a new population that will be modified by a mutation that inserts a new command in the command list of a PrimaryShape
	int middlePercent; //percent of insertion mutations that insert a command in the middle of the PrimaryShape's command list
	int endPercent; //" " end of a PrimaryShape's command list
	
	int	changeMutationPercent; //percent (out of 100) of individuals in a new population that will be modified by a mutation that changes an existing command in a PrimaryShape's command list
	int ringPercent; //percent of change mutations that alter a command's PERIng (does not apply to openings?)
	int paramPercent; //percent of change mutations that alter parameters of a command, such as the number of degrees of a rotation
	int completePercent; //percent of change mutations that completely change a command

	int deletePercent; //percent of individuals that will be modified by a mutation that deletes a command from the list of one shape

	int	crossoverPercent; //percent (out of 100) of individuals in a new population that will be created by crossing over the command lists of two individuals in the current population
	int minNumberOfCrossingPoints; //minimum number of points at which the source of the new individual's transformation list is changed
	int maxNumberOfCrossingPoints; //maximum number of points "" ""		
	static int minCrossoverSize; //DEFINE THIS BY WRITING CODE
	static int maxCrossoverSize; //DEFINE THIS BY WRITING CODE
	
	int	reproductionPercent; //percent (out of 100) of individuals in a new population that will consist of unmodified members of the previous generation
	int	elitism; //number of individuals to be passed through unchanged
	int	eliteCopies; //number of copies of each elite individual to be passed into the next generation

	// Mutation Constants
	static int translationPercent;
	static int rotationPercent;
	static int scalePercent;
	static int reflectionPercent;
	static int openingPercent;

	// MAXES AND MINS
	// ZWidening constants
	static double ZWideningMin;
	static double ZWideningMax;

	// ZHeight constants
	static double ZHeightMin;
	static double ZHeightMax;

	// ZStart constants
	static double ZStartMin;
	static double ZStartMax;

	// Reporting variables:
	int numElited;
	int numParamMutated;
	int numPETyped; //numer of param mutations
	int numZWidened; //" " ZWidening
	int numshapeModed; //" " shapeMode (1 - 45 how intersections are handled)
	int numZHeighted; //" " ZHeight
	int numZStarted; //" " ZStart
	
	int	numInsertionMutated; //numer of individuals in a new population that will be modified by a mutation that inserts a new command in the command list of a PrimaryShape
	int numMiddled; //numer of insertion mutations that insert a command in the middle of the PrimaryShape's command list
	int numEnded; //" " end of a PrimaryShape's command list
	
	int	numChangeMutated; //numer of individuals in a new population that will be modified by a mutation that changes an existing command in a PrimaryShape's command list
	int numRinged; //numer of mutations that alter a command's PERIng (does not apply to openings?)
	int numParamed; //numer of change mutations that alter parameters of a command, such as the number of degrees of a rotation
	int numCompleted; //numer of change mutations that completely change a command

	int numCrossedOver; // number of crossovers
	int numReproduced; // number of reproductions
	int numDeleted;  // number of deletion mutations

	bool bIsSeeded;

	MSelectionList* _selectionList;

	GeneticEngine* initializeEngine();
// Interface stuff (a little sad)
	PopulationProperties*	popPropsPage;

public:
	void makeAllImpressionsInvalid();
	static GeneticEngine* geneticEngine;

// Settables inheritance
protected:

	virtual void reportSettings();
	virtual void receiveSettings();


};



#endif