//
// Copyright (C) 2001 Simon Greenwold. You are allowed to use or modify the code, but this note must be kept intact.
// 
// Genotype.cpp: implementation of the Genotype class.
//
//////////////////////////////////////////////////////////////////////
#include "stdafx.h"
#include <maya/MGlobal.h>
#include "Genotype.h"
#include "GeneticUtility.h" //Moved from Genotype.h
#include "Constants.h"
#include "PrimaryShape.h"
#include <iostream>
#include "ZoneTypes.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

Genotype::Genotype()
{
	shapes = new std::vector<PrimaryShape*>();
}


Genotype::Genotype(const std::vector<MayaInfo*>& originalCurves, const std::vector<string>& layerNames)
{
	shapes = new std::vector<PrimaryShape*>(originalCurves.size());
	for (int i = 0; i < originalCurves.size(); i++)
		(*shapes)[i] = new PrimaryShape(originalCurves[i], layerNames[i]);
}

Genotype::Genotype(const Genotype& oldGenotype) {

	shapes = new std::vector<PrimaryShape*>(oldGenotype.getShapes()->size());

	for(int i = 0; i < shapes->size(); i++) {
		(*shapes)[i] = new PrimaryShape(*(*(oldGenotype.getShapes()))[i]);
	}
}


Genotype::~Genotype()
{
	clearShapes();
}

void Genotype::clearShapes() {

	if (TRACE_LEVEL > 9) cout << "Into clear shapes." << endl;

	//shapes must be deleted
	for (int i = 0; i < shapes->size(); i++) {
		delete (*shapes)[i];
	}


	delete shapes;
}

void Genotype::setShapes(std::vector<PrimaryShape*>* newShapes) {
	clearShapes();
	shapes = newShapes;
}


void Genotype::mutateParam(ParamType typeOfParam) {
	int mutationTarget = getShapeToMutate();
	if(mutationTarget!=-1)	
		(*shapes)[mutationTarget]->mutateParams(typeOfParam);
}

void Genotype::mutateInsertion(InsertionType typeOfInsertion) {
	int mutationTarget = getShapeToMutate();
	if(mutationTarget!=-1)
		(*shapes)[mutationTarget]->mutateInsertion(typeOfInsertion);
}

void Genotype::mutateDeletion() {
	int mutationTarget = getShapeToMutate();
	if(mutationTarget!=-1)
		(*shapes)[mutationTarget]->mutateDeletion();
}
void Genotype::mutateChange(ChangeType typeOfChange) {
	int mutationTarget = getShapeToMutate();
	if(mutationTarget!=-1)
		(*shapes)[mutationTarget]->mutateChange(typeOfChange);
}

void Genotype::crossover(int numberOfCrossingPoints, Design* crossoverTemplate) {
	int mutationTarget = getShapeToMutate();
	if(mutationTarget!=-1)
		(*shapes)[mutationTarget]->crossover(numberOfCrossingPoints, crossoverTemplate);
}

void Genotype::randomize(int minNumberOfStartingTransformations, int maxNumberOfStartingTransformations) {

	int numberOfTransformations;

	for(int i = 0; i < shapes->size(); i++) {
		(*shapes)[i]->mutateParams(PETYPE); //randomly set parameters of each shape
		(*shapes)[i]->mutateParams(ZWIDENING);
		(*shapes)[i]->mutateParams(SHAPEMODE);
		(*shapes)[i]->mutateParams(ZHEIGHT);
		(*shapes)[i]->mutateParams(ZSTART);

		(*shapes)[i]->clearOperations();  // Get rid of any existing operations in the list
		numberOfTransformations = randomRange(minNumberOfStartingTransformations, maxNumberOfStartingTransformations);
		for(int j = 0; j < numberOfTransformations; j++) {
			(*shapes)[i]->mutateInsertion(END); //add the appropriate number of random transformations to each shape
		}
	}
}

const Genotype& Genotype::operator=(const Genotype& oldGenotype) {

	clearShapes();

	shapes = new std::vector<PrimaryShape*>(oldGenotype.getShapes()->size());

	for(int i = 0; i < shapes->size(); i++) {
		(*shapes)[i] = new PrimaryShape(*(*(oldGenotype.getShapes()))[i]);
	}
	return oldGenotype;
}

void Genotype::reduceNumZoneTypes(int newNumZoneTypes)
{
	for(int i = 0; i < shapes->size(); i++) {
		(*shapes)[i]->reduceNumZoneTypes(newNumZoneTypes);
	}
}

int Genotype::getShapeToMutate()
{
	vector<PrimaryShape*> copyOfShapes(*shapes);	//Make a copy of the shape list
	
	vector<bool> immutables;	//this vector contains the shapes that we tried to mutate but was not allowed to.
	immutables.resize(copyOfShapes.size());
	for(int i = 0; i < immutables.size(); i++)	
		immutables[i] = false;	//initialize the vector

	int offset = 0;

	while(!copyOfShapes.empty())	{
		int mutatenr = safeRandom(copyOfShapes.size() - 1);
		if(safeRandom(99) < ZoneTypes::zoneTypes->getMutationRate(copyOfShapes[mutatenr]->getOriginalPEType().getType()))	{
			for(int j = 0; j < mutatenr; j++)	{
				if(immutables[j])	
					offset++;
			}
			return mutatenr + offset;
		}
		else	{
			vector<PrimaryShape*>::iterator it = copyOfShapes.begin();	//Delete mutatenr from the vector
			for(int k = 0; k < mutatenr; k++, it++);
			copyOfShapes.erase(it);
			immutables[mutatenr] = true;	//Store the number in another list
		}
	}
	return -1;	//Could not find any mutable shape
}
