//
// Copyright (C) 2001 Simon Greenwold. You are allowed to use or modify the code, but this note must be kept intact.
// 
// PrimaryShape.cpp: implementation of the PrimaryShape class.
//
//////////////////////////////////////////////////////////////////////
#include "stdafx.h"
#include <maya/MGlobal.h>

#include "PrimaryShape.h"

#include "Constants.h"
#include "Utilities.h"

#include <maya/MDagPath.h>
#include <maya/MFnTransform.h>
#include <maya/MGlobal.h>
#include <maya/NewMVector.h>
#include <maya/MSelectionList.h>
#include <maya/MItSelectionList.h>
#include <maya/MTransformationMatrix.h>
#include <maya/MFnDagNode.h>
#include <maya/MDagModifier.h>
#include "MELUtilities.h"

#include <sstream>
#include "GeneticUtility.h"
#include "Design.h"
#include "Genotype.h"
#include "Random.h"
#include "Transformation.h"
#include "Translation.h"
#include "Rotation.h"
#include "Scale.h"
#include "Reflection.h"
#include "Opening.h"
#include "GeneticEngine.h"
#include "MayaUtilities.h"
#include <iostream>

using namespace std;


//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

PrimaryShape::PrimaryShape(MayaInfo* original, const string& originalLayer) 
: Shape()
{
	originalCurve = original;
	_originalLayer = originalLayer;
}

PrimaryShape::PrimaryShape(const PrimaryShape& oldShape) {

	originalAssignment = oldShape.originalAssignment;
	originalZWidening = oldShape.originalZWidening;
	originalShapeMode = oldShape.originalShapeMode;
	originalZHeight = oldShape.originalZHeight; 
	originalZStart = oldShape.originalZStart;
	
	assignment = oldShape.assignment;
	ZWidening = oldShape.ZWidening;
	shapeMode = oldShape.shapeMode;
	ZHeight = oldShape.ZHeight; 
	ZStart = oldShape.ZStart;
	ZStartScale = oldShape.ZStartScale;
	hasCut = oldShape.hasCut;
	cutStart = oldShape.cutStart;
	cutStop = oldShape.cutStop;

	mayaShape = new MayaInfo();
	mayaShape->setName(oldShape.mayaShape->name());

	originalCurve = oldShape.originalCurve;
	_originalLayer = oldShape._originalLayer;

	list<Command*>::const_iterator position = oldShape.getOperations().begin();
	for(int i = 0; i < oldShape.getOperations().size(); i++) {
		operations.push_back((*(position++))->clone()); // Dereference the iterator, clone it, find its address, and push that onto the operations list.  Increment the iterator 
	}
	eraseOnDelete = true;
}

PrimaryShape::~PrimaryShape()
{
	clearOperations();  //operations do have to be deleted

}


void PrimaryShape::clearOperations() {

	while(operations.size() > 0) {
		delete (*operations.begin());
		operations.pop_front();
	}

}

//Genetic Operations

void PrimaryShape::mutateParams(ParamType typeOfParam) {
	switch(typeOfParam) { //PETYPE, ZWIDENING, SHAPEMODE, ZHEIGHT, ZSTART
	case PETYPE:
		originalAssignment.mutate();
		break;
	case ZWIDENING:
		originalZWidening = Random::genrand() * (GeneticEngine::getZWideningMax() - GeneticEngine::getZWideningMin()) + GeneticEngine::getZWideningMin(); 
		break;
	case SHAPEMODE:
		originalShapeMode = randomRange(1,45);
		break;
	case ZHEIGHT:
		originalZHeight = Random::genrand() * (GeneticEngine::getZHeightMax() - GeneticEngine::getZHeightMin()) + GeneticEngine::getZHeightMin();
		break;
	case ZSTART:
		originalZStart = Random::genrand() * (GeneticEngine::getZStartMax() - GeneticEngine::getZStartMin()) + GeneticEngine::getZStartMin();
		break;
	}
}

Command* PrimaryShape::createRandomCommand() {

	Command *newCommand;

	int transformationType = safeRandom(99);

	if (transformationType < GeneticEngine::getTranslationPercent()) {
		newCommand = new Translation();
	}
	else if((transformationType -= GeneticEngine::getTranslationPercent()) < GeneticEngine::getRotationPercent()) {
		newCommand = new Rotation();
	}
	else if((transformationType -= GeneticEngine::getRotationPercent()) < GeneticEngine::getScalePercent()) {
		newCommand = new Scale();
	}
	else if((transformationType -= GeneticEngine::getScalePercent()) < GeneticEngine::getReflectionPercent()) {
		newCommand = new Reflection();
	}
	else if((transformationType -= GeneticEngine::getReflectionPercent()) < GeneticEngine::getOpeningPercent()) {
		newCommand = new Opening();
	}
	else {
		//ERROR - this should not happen
		printf("Error in createRandomCommand!");
		newCommand = new Translation();
	}

	newCommand->randomize();

	return newCommand;
}

void PrimaryShape::mutateInsertion(InsertionType typeOfInsertion) {
	//create new transformation;
	Command* newCommand = createRandomCommand();	
	//insert new transformation

	list<Command*>::iterator position = operations.begin();	
	int i = 0;
	int insertionPoint = 0;
	switch(typeOfInsertion) {
	case MIDDLE:
		insertionPoint = safeRandom(operations.size());
		for(i = 0; i < insertionPoint; i++) {
			position++;
		}
		operations.insert(position, newCommand);
		break;
	case END:
		operations.insert(operations.end(), newCommand);
	}
}

void PrimaryShape::mutateDeletion() {

	if (operations.size() == 0) return;

	list<Command*>::iterator position = operations.begin();
	int i = 0;
	int deletionPoint = safeRandom(operations.size() - 1);
	for(i = 0; i < deletionPoint; i++, position++);
	delete *position;
	operations.erase(position);
}

void PrimaryShape::mutateChange(ChangeType typeOfChange) {  // Makes a change to an existing command in a shape's list

	//RING, PARAM, COMPLETE
	if (operations.size() == 0) return;
	int changedCommand = safeRandom(operations.size() - 1);
	std::list<Command*>::iterator position = operations.begin();
	for(int i = 0; i < changedCommand; position++, i++) ; //move position iterator to the correct element
	//The above process would be MUCH faster if a vector was used instead of a list - seriously consider this change

	switch(typeOfChange) {
	case RING: //if the selected command is a Transformation, assign it a new PERing.  If the selected command is an opening (a non-Transformation command, do nothing)
		(*position)->randomizePERing();
		break;

	case PARAM: //mutate the parameters of the selected command
		(*position)->mutate();
		break;

	case COMPLETE:
		delete (*position);
		(*position) = createRandomCommand();
		break;
	}

}

void PrimaryShape::crossover(int numberOfCrossingPoints, Design* crossoverTemplate) {

	const std::vector<PrimaryShape*>* templateShapes = crossoverTemplate->getGenome()->getShapes();
	PrimaryShape *templateShape = (*templateShapes)[safeRandom(templateShapes->size() - 1)];
	//The ideal method of selecting crossover points seems less than obvious - I have choosen a method that is both fairly reasonable yet easy to program
	//REJECTED CONCEPT - Crossovers are not allowed to overlap.  New crossoverpoints are choosen based upon a list of options that does not include previously selected points.  If a crossover cannot fit in the choosen space, it is ignored
	//Instead - The prior idea carried the biological metaphor too far.  The idea of crossover in this case is merely to introduce successful code from one individual into another.  Since there are not "genes" or corresponding functional units between individuals, it is irrelevant where the swapped code is placed, whether swapped code overlaps with other swapped code, or even the number of consecutive units of original code maintained after cross-over.  
	//Thus, for the sake of simplicity, code of appropriate lengths is taken at random from the template and replaces random code in the target, regardless of overlap.  The length of the replacing code and the length of the replaced code need not be equal  
	
	int crossoverPointTarget;
	int crossoverLengthTarget;
	int crossoverPointTemplate;
	int crossoverLengthTemplate;
		
	for(int i = 0; i < numberOfCrossingPoints; i++) {

		crossoverPointTarget = safeRandom(operations.size() - 1); 
		crossoverLengthTarget = randomRange(GeneticEngine::getMinCrossoverSize(), GeneticEngine::getMaxCrossoverSize());
		if(crossoverPointTarget + crossoverLengthTarget >= operations.size())
			crossoverLengthTarget = operations.size() - crossoverPointTarget; //assure that the crossover does not run over the length of the Command* list - truncate crossover if necessary
		
		crossoverPointTemplate = safeRandom(templateShape->getOperations().size());
		crossoverLengthTemplate = randomRange(GeneticEngine::getMinCrossoverSize(), GeneticEngine::getMaxCrossoverSize());
		if(crossoverPointTemplate + crossoverLengthTemplate >= templateShape->getOperations().size())
			crossoverLengthTemplate = templateShape->getOperations().size() - crossoverPointTemplate; //assure that crossover does not run over - see above
		
		for(std::list<Command*>::iterator deleteIterator = addIterator(operations.begin(), crossoverPointTarget); deleteIterator != addIterator(operations.begin(), crossoverPointTarget + crossoverLengthTarget); deleteIterator++) {
			delete (*(deleteIterator)); //delete each command before removing it from operations (since the operations list holds only pointers
		}
		operations.erase(addIterator(operations.begin(), crossoverPointTarget), addIterator(operations.begin(), crossoverPointTarget + crossoverLengthTarget)); //assumes that this function (.erase()) erases from the first iterator to the element BEFORE the last iterator - may instead erase the last mentioned element also
		
		for(int j = 0; j < crossoverLengthTemplate; j++) {
			Command* newCommand = (*(addIterator(templateShape->getOperations().begin(), crossoverPointTemplate + j)))->clone();
			operations.insert(addIterator(operations.begin(), crossoverPointTarget + j), newCommand);
		}
	}
		
}

std::list<Command*>::iterator PrimaryShape::addIterator(std::list<Command*>::iterator startingPoint, int addend) {
	for(int i = 0; i < addend; i++) {
		startingPoint++;
	}
	return startingPoint;
}

std::list<Command*>::const_iterator PrimaryShape::addIterator(std::list<Command*>::const_iterator startingPoint, int addend) {
	for(int i = 0; i < addend; i++) {
		startingPoint++;
	}
	return startingPoint;
}

// TRANSFORMATIONS

bool PrimaryShape::scale(double x, double y, double z) {
	MStatus				status;
	double				scalings[3];

	if (!mayaShape->transform()) {
		cout << "Error: No shape named \"" << mayaShape->name() << "\"." << endl;
		return false;
	}

	status = mayaShape->transform()->getScale(scalings);
	if (!status) {
		status.perror("MFnTransformNode getScale");
		return false;
	}

	scalings[0] *= x;
	scalings[2] *= y;
	ZHeight *= z;

	status = mayaShape->transform()->setScale(scalings);
	if (!status) {
		status.perror("MFnTransformNode setScale");
		return false;
	}

//	cout << "Shape \"" << mayaShape->name() << "\" scaled (" << x << ", " << y << ", " << z << ")." << endl;

	return true;
}

bool PrimaryShape::translate(double x, double y, double z) {
	MStatus				status;
	MVector				translations;

	if (!mayaShape->transform()) {
		cout << "No shape named " << mayaShape->name() << endl;
		return false;
	}

	translations[0] = x;
	translations[1] = 0;
	translations[2] = y;
	ZStart += z;
//	ZHeight += z;   // Instead of translating up, let's pull the top up...

	status = mayaShape->transform()->translateBy(translations, MSpace::kTransform);
	if (!status) {
		status.perror("MFnTransformNode translateBy");
		return false;
	}

//	cout << "Shape \"" << mayaShape->name() << "\" translated (" << x << ", " << y << ", " << z << ")." << endl;

	return true;
}

bool PrimaryShape::opening(double start, double stop) {

	hasCut = true;
	if (start < cutStart) cutStart = start;
	if (stop > cutStop) cutStop = stop;
	return true;
}

bool PrimaryShape::rotate(double degrees) {
	MStatus				status;
	double				rotation[3];

	if (!mayaShape->transform()) {
		cout << "No shape named " << mayaShape->name() << endl;
		return false;
	}

	rotation[0] = 0;
	rotation[1] = (degrees / 180.0) * PI;
	rotation[2] = 0;

	status = mayaShape->transform()->rotateBy ( rotation, MTransformationMatrix::kXYZ, MSpace::kTransform);
	if (!status) {
		status.perror("MFnTransformNode rotateBy");
		return false;
	}

//	cout << "Shape \"" << mayaShape->name() << "\" rotated by (" << degrees << ")." << endl;

	return true;
}

bool	PrimaryShape::initializeInterpret() {

	erase();						// Make sure there aren't any objects hanging around
	string newName;

	if (duplicateCurve(originalCurve, newName))	{	// Make a copy of the original curve for this to refer to
		mayaShape->setName(newName);
		cleanupList.push_back(newName);
	}	
	nextCommand = operations.begin();


									// Set the operable parameters to their original values
	assignment = originalAssignment;
	ZWidening = originalZWidening;
	shapeMode = originalShapeMode;
	ZHeight = originalZHeight;
	ZStart = originalZStart;
	hasCut = false;
	cutStart = 1;
	cutStop = 0;

	return true;
}

void	PrimaryShape::runCommand(Command* theCommand) {

	theCommand->run(this);
}

bool	PrimaryShape::runNextCommand() {

	if (nextCommand != operations.end()) {
		runCommand(*nextCommand);
		nextCommand++;
		return true;
	}
	else
		return false;
}




const PrimaryShape& PrimaryShape::operator=(const PrimaryShape& oldShape) {
	originalAssignment = oldShape.originalAssignment;
	originalZWidening = oldShape.originalZWidening;
	originalShapeMode = oldShape.originalShapeMode;
	originalZHeight = oldShape.originalZHeight; 
	originalZStart = oldShape.originalZStart;
	
	assignment = oldShape.assignment;
	ZWidening = oldShape.ZWidening;
	shapeMode = oldShape.shapeMode;
	ZHeight = oldShape.ZHeight; 
	ZStart = oldShape.ZStart;
	ZStartScale = oldShape.ZStartScale;
	hasCut = oldShape.hasCut;
	cutStart = oldShape.cutStart;
	cutStop = oldShape.cutStop;

	//the function of the Maya representation variables is not clear and so they are not copied
	if (mayaShape != NULL) delete mayaShape;
	mayaShape = new MayaInfo();
	originalCurve = oldShape.originalCurve;

	for(list<Command*>::const_iterator originalPosition = operations.begin(); originalPosition != operations.end(); originalPosition++) { //delete old operations before copying new ones
		delete (*(originalPosition));
	}
	const list<Command*> tempOperations = oldShape.getOperations();
	operations.resize(0); //since this is a list, eliminate all previous elements and build up from scratch
	for(list<Command*>::const_iterator position = tempOperations.begin();  position != tempOperations.end(); position++) {
		operations.push_back((*position)->clone()); //push a copy of the relevant command onto the list 
	}
	cleanupList.clear();
	eraseOnDelete = true;
	return oldShape;
}

Shape*	PrimaryShape::fullDuplicate()
{
	PrimaryShape* newShape = new PrimaryShape(*this);
	string newName;
	if (duplicateCurve(mayaShape, newName))	{	// Make a copy of the original curve for this to refer to
		mayaShape->setName(newName);
		cleanupList.push_back(newName);
	}
	return newShape;
}

void PrimaryShape::appendOperation(Command* newCommand)
{
	operations.push_back(newCommand);
}


void PrimaryShape::insertOperationBefore(Command* insertionPoint, Command* newCommand)
{
	for (list<Command*>::iterator i = operations.begin(); i != operations.end(); i++) {
		if (insertionPoint == *i) {
			operations.insert(i, newCommand);
			return;
		}
	}
}

void PrimaryShape::removeOperation(Command* oldCommand)
{
	for (list<Command*>::iterator i = operations.begin(); i != operations.end(); i++) {
		if (oldCommand == *i) {

			delete oldCommand;
			operations.erase(i);
			return;
		}
	}
}

void PrimaryShape::reduceNumZoneTypes(int newNumZoneTypes)
{
	if (assignment.getType() >= newNumZoneTypes) assignment.mutate();
	if (originalAssignment.getType() >= newNumZoneTypes) originalAssignment.mutate();


	for (list<Command*>::iterator i = operations.begin(); i != operations.end(); i++)
		(*i)->randomizePERing();
}
