//
// Copyright (C) 2001 Simon Greenwold. You are allowed to use or modify the code, but this note must be kept intact.
// 
// ZoneTypes.cpp: implementation of the ZoneTypes class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include <maya/MGlobal.h>
#include "MFCPlugin.h"
#include "ZoneTypes.h"
#include "Settings.h"
#include "Constants.h"
#include <iostream>
#include <sstream>
#include "Layer.h"
#include "PrimaryShape.h"
#include "PEType.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

ZoneTypes* ZoneTypes::zoneTypes;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

ZoneTypes::ZoneTypes()
{
	_shaderGroups.empty();
	_geneticEngine = NULL;
	_agentControl = NULL;
	zoneTypes = this;
}

void ZoneTypes::initializeZoneTypes()
{
	receiveSettings();
}

ZoneTypes::~ZoneTypes()
{

}

void ZoneTypes::setEngine(GeneticEngine* geneticEngine)
{
	_geneticEngine = geneticEngine;
}

void ZoneTypes::setAgentControl(AgentControl* agentControl)
{
	_agentControl = agentControl;
}

bool ZoneTypes::validShader(const string& testShader)
{
	for (list<string>::iterator i = _shaderGroups.begin(); i != _shaderGroups.end(); i++)
		if (*i == testShader) return true;
	return false;
}

void ZoneTypes::addShader(const string& shader)
{
	_shaderGroups.push_back(shader);
}

void ZoneTypes::setNumZoneTypes(int newNumZoneTypes)	// Do the work (nasty)
{
	if (newNumZoneTypes == _numZoneTypes) return;

	if (_geneticEngine)
		if (newNumZoneTypes < _numZoneTypes) {
			_numZoneTypes = newNumZoneTypes;
			_geneticEngine->reduceNumZoneTypes(newNumZoneTypes);
		}

	_numZoneTypes = newNumZoneTypes;
	_zoneNames.resize(_numZoneTypes);
	_zoneShaders.resize(_numZoneTypes);
	_mutateInOut.resize(_numZoneTypes);
	_mutationRate.resize(_numZoneTypes);
}

void ZoneTypes::deleteZoneType(int typeNum)
{
	for (int i = typeNum + 1; i < _numZoneTypes; i++) {
		_zoneNames[i - 1] = _zoneNames[i];
		_zoneShaders[i - 1] = _zoneShaders[i];
		_mutateInOut[i - 1] = _mutateInOut[i];
		_mutationRate[i - 1] = _mutationRate[i];
	}
	setNumZoneTypes(_numZoneTypes - 1);

	
	if (_agentControl) _agentControl->removeZoneType(typeNum);
}

void ZoneTypes::addZoneType(const string& zoneName, const string& zoneShader, int typeNum)
{
	setNumZoneTypes(_numZoneTypes + 1);
	typeNum++;

	for (int i = _numZoneTypes - 1; i > typeNum; i--) {
		_zoneNames[i] = _zoneNames[i - 1];
		_zoneShaders[i] = _zoneShaders[i - 1];
		_mutateInOut[i] = _mutateInOut[i - 1];
		_mutationRate[i] = _mutationRate[i - 1];
	}
	_zoneNames[typeNum] = zoneName;
	_zoneShaders[typeNum] = zoneShader;

	_mutateInOut[typeNum] = false;
	_mutationRate[typeNum] = MUTATIONRATE;

	if (_agentControl) _agentControl->addZoneType(typeNum - 1);
}

void ZoneTypes::receiveSettings()
{
	Settings::zoneTypesInitialized = true;
	int tempInt, tempBool;
	ostringstream tempString;

	if (! Settings::settings->getSetting("ZoneTypes", "numZoneTypes", tempInt))
		tempInt = PETYPE_MAX - PETYPE_MIN + 1;

	setNumZoneTypes(tempInt);

	list<string>::iterator j = _shaderGroups.begin();


	for (int i = 0; i < _numZoneTypes; i++) {
		if (! Settings::settings->getSetting("ZoneTypes", "zoneTypeName", _zoneNames[i], i)) {
			tempString.str("");
			tempString << "ZoneType " << i + 1;
			_zoneNames[i] = tempString.str();
		}

		if (! Settings::settings->getSetting("ZoneTypes", "zoneTypeShader", _zoneShaders[i], i)) {
			tempString.str("");
			tempString << "shader" << i + 1 << "SG";
			_zoneShaders[i] = tempString.str();
		}
		if (! validShader(_zoneShaders[i])) {
			_zoneShaders[i] = *j;
			j++;
			if (j == _shaderGroups.end()) j = _shaderGroups.begin();
		}

		if (Settings::settings->getSetting("ZoneTypes", "zoneTypeMutatable", tempBool, i)) {
			_mutateInOut[i] = tempBool;

		}
		else {
			_mutateInOut[i] = false;
		}
		if (! Settings::settings->getSetting("ZoneTypes", "zoneTypeMutationRate", _mutationRate[i], i)) {
			_mutationRate[i] = MUTATIONRATE;
		}
		
	}
	if (_agentControl) {
		_agentControl->setNumZoneTypes(tempInt);
	}
	checkLayerNames();
}

void ZoneTypes::reportSettings()
{
	Settings::settings->setSetting("ZoneTypes", "numZoneTypes", _numZoneTypes);
	for (int i = 0; i < _numZoneTypes; i++) {
		Settings::settings->setSetting("ZoneTypes", "zoneTypeName", _zoneNames[i], i);
		Settings::settings->setSetting("ZoneTypes", "zoneTypeShader", _zoneShaders[i], i);
		Settings::settings->setSetting("ZoneTypes", "zoneTypeMutatable", _mutateInOut[i], i);
		Settings::settings->setSetting("ZoneTypes", "zoneTypeMutationRate", _mutationRate[i], i);
	}
}

void ZoneTypes::checkLayerNames()
{
	//Check if there is a layer with the same name
	string zoneName, layerName;
	int zoneType;
	for(int i=0; i < getNumZoneTypes(); i++)	{
		zoneName = getZoneName(i);
		for(int j=0; j < Layer::layer->numberOfLayers(); j++)	{
			layerName = Layer::layer->getLayer(j);
			if(Layer::layer->getLayer(j)==zoneName)	{
				Layer::layer->bindLayerToZoneType(j);
				//Loop through the population
				for(int k=0; k < GeneticEngine::geneticEngine->getPopulationSize(); k++)	{
					Phenotype* phenome = &(GeneticEngine::geneticEngine->getPopulation()->getDesign(k)->getPhenome());
					for(list<Shape*>::iterator l=phenome->getShapes().begin(); l != phenome->getShapes().end(); l++)	{
						if((*l)->isPrimary())	{
							if(((PrimaryShape*)*l)->getOriginalLayer() == layerName)	{
								((PrimaryShape*)*l)->setOriginalPEType(i);	//Switch ZoneType for all curves in that layer
							}
							else if(((PrimaryShape*)*l)->getOriginalPEType().getType() == i)	{
								((PrimaryShape*)*l)->setOriginalPEType(newZoneType(i)); 	//Switch ZoneType for all curves that belong to another layer
							}
						}
					}
				}
			}
		}
	}
}

//Returns a zoneType that is not equal to i
int ZoneTypes::newZoneType(int i)
{
	int numZoneTypes = getNumZoneTypes();
	int zoneTypeToTry, numZoneTypesLeftToTry = numZoneTypes;
	
	zoneTypeToTry = safeRandom(numZoneTypes - 1);
	while(numZoneTypesLeftToTry > 0)	{
		if(!Layer::layer->LayerBoundToZoneType(zoneTypeToTry))	{
			return zoneTypeToTry;
		}
		else	{
			++zoneTypeToTry % numZoneTypes;
			numZoneTypesLeftToTry--;
		}
	}
	return i;	//Failed to find an OK ZoneType, return the same
}